#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <boost/scoped_array.hpp>
//PE library header file 
#include <pe_32_64.h>
//LZO1Z999 algorithm header file
#include "../../lzo-2.06/include/lzo/lzo1z.h"
//Our structures header file
#include "structs.h"
//Header file with unpacker parameters
#include "../unpacker/parameters.h"
//Unpacker body (autogenerated)
#include "unpacker.h"

//PE and LZO libraries linking directives
#ifndef _M_X64
#ifdef _DEBUG
#pragma comment(lib, "../../Debug/pe_lib.lib")
#pragma comment(lib, "../Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../Release/pe_lib.lib")
#pragma comment(lib, "../Release/lzo-2.06.lib")
#endif
#else
#ifdef _DEBUG
#pragma comment(lib, "../../x64/Debug/pe_lib.lib")
#pragma comment(lib, "../x64/Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../x64/Release/pe_lib.lib")
#pragma comment(lib, "../x64/Release/lzo-2.06.lib")
#endif
#endif

int main(int argc, char* argv[])
{
	//Tell the user how to use our packer
	//No packing options at this step
	//You just need to pass a name of the file to pack
	//via command line
	if(argc != 2)
	{
		std::cout << "Usage: simple_pe_packer.exe PE_FILE" << std::endl;
		return 0;
	}

	//Open a file - its name is stored in argv array at index 1
	std::ifstream file(argv[1], std::ios::in | std::ios::binary);
	if(!file)
	{
		//If the file was not opened successfully - notify user and exit with error
		std::cout << "Cannot open " << argv[1] << std::endl;
		return -1;
	}

	try
	{
		//Try to open the file as 32-bit PE file
		//Last two arguments are false, because we don't need them
		//"raw" file bound import data and 
		//"raw" debug information data
		//They are not used while packing, so we don't load these data
		pe32 image(file, false, false);

		//Check if .NET image was passed
		if(image.is_dotnet())
		{
			std::cout << ".NEt image cannot be packed!" << std::endl;
			return -1;
		}

		//Calculate file sections entropy to make sure that file was not packed
		{
			std::cout << "Entropy of sections: ";
			double entropy = image.calculate_entropy();
			std::cout << entropy << std::endl;
		//There is an opinion,
		//that PE files have normal entropy below 6.8
		//If it has more, than file is most possibly compressed
		//So (for now) we will not pack files
		//with high entropy, this makes little sense
			if(entropy > 6.8)
			{
				std::cout << "File has already been packed!" << std::endl;
				return -1;
			}
		}

		//Initialize LZO compression library
		if(lzo_init() != LZO_E_OK)
		{
			std::cout << "Error initializing LZO library" << std::endl;
			return -1;
		}
		
		std::cout << "Reading sections..." << std::endl;

		//Get PE file sections list
		const pe_base::section_list& sections = image.get_image_sections();
		if(sections.empty())
		{
			//If file has no sections, we have nothing to pack
			std::cout << "File has no sections!" << std::endl;
			return -1;
		}

		//PE file basic information structure
		packed_file_info basic_info = {0};
		//Get and save original section count
		basic_info.number_of_sections = sections.size();

		//String to store each section packed_section structure one-by-one
		std::string packed_sections_info;

		{
			//Allocate required memory for these structures
			packed_sections_info.resize(sections.size() * sizeof(packed_section));

			//Raw data of all sections, read from the file and assembled together
			std::string raw_section_data;
			//Current section index
			unsigned long current_section = 0;

			//List all sections
			for(pe_base::section_list::const_iterator it = sections.begin(); it != sections.end(); ++it, ++current_section)
			{
				//Next section reference
				const pe_base::section& s = *it;

				{
					//Create section information structure
					//and fill it
					packed_section& info
						= reinterpret_cast<packed_section&>(packed_sections_info[current_section * sizeof(packed_section)]);

					//Section characteristics
					info.characteristics = s.get_characteristics();
					//File data pointer
					info.pointer_to_raw_data = s.get_pointer_to_raw_data();
					//File data size
					info.size_of_raw_data = s.get_size_of_raw_data();
					//Relative virtual section address
					info.virtual_address = s.get_virtual_address();
					//Virtual section address
					info.virtual_size = s.get_virtual_size();

					//Copy section name (8 symbols maximum)
					memset(info.name, 0, sizeof(info.name));
					memcpy(info.name, s.get_name().c_str(), s.get_name().length());
				}

				//If section is empty, switch no the next one
				if(s.get_raw_data().empty())
					continue;

				//If it is not empty - copy its data to a string
				//containing all sections data
				raw_section_data += s.get_raw_data();
			}

			//If all sections are empty, there is nothing to pack!
			if(raw_section_data.empty())
			{
				std::cout << "All sections of PE file are empty!" << std::endl;
				return -1;
			}

			//We will pack both buffers assembled together
			//(see below)
			packed_sections_info += raw_section_data;
		}

		//New section
		pe_base::section new_section;
		//Name - .rsrc (see description below)
		new_section.set_name(".rsrc");
		//Available for reading, writing, execution
		new_section.readable(true).writeable(true).executable(true);
		//Reference to section raw data
		std::string& out_buf = new_section.get_raw_data();

		//Create smart pointer
		//and allocate memory required by LZO algorithm
		//Smart pointer will release this memory
		//automatically if needed
		//We use lzo_align_t type to
		//align memory as required
		//(from LZO documentation)
		boost::scoped_array<lzo_align_t> work_memory(new lzo_align_t[LZO1Z_999_MEM_COMPRESS]);
		
		//Unpacked data length
		lzo_uint src_length = packed_sections_info.size();
		//Save it to our file information structure
		basic_info.size_of_unpacked_data = src_length;
		
		//Packed data length
		//(unknown yet)
		lzo_uint out_length = 0;
		
		//Compressed data buffer
		//(length based on LZO documentation recommendations)
		out_buf.resize(src_length + src_length / 16 + 64 + 3);

		//Perform data compression
		std::cout << "Packing data..." << std::endl;
		if(LZO_E_OK !=
			lzo1z_999_compress(reinterpret_cast<const unsigned char*>(packed_sections_info.data()),
			src_length,
			reinterpret_cast<unsigned char*>(&out_buf[0]),
			&out_length,
			work_memory.get())
			)
		{
			//If something goes wrong, exit
			std::cout << "Error compressing data!" << std::endl;
			return -1;
		}
		
		std::cout << "Packing complete..." << std::endl;

		//Store packed data length in our structure
		basic_info.size_of_packed_data = out_length;
		//Resize output buffer with compressed data by
		//resulting compressed data length
		//which is now known
		out_buf.resize(out_length);
		//Assemble the buffer, this will be
		//resulting data of our new section
		out_buf =
			//basic_info structure data
			std::string(reinterpret_cast<const char*>(&basic_info), sizeof(basic_info))
			//Output buffer
			+ out_buf;

		//Check if file became really smaller
		if(out_buf.size() >= src_length)
		{
			std::cout << "File is incompressible!" << std::endl;
			return -1;
		}


		{
			//Get reference to first existing 
			//PE file section
			const pe_base::section& first_section = image.get_image_sections().front();
			//Set virtual address of section to be added (see below)
			new_section.set_virtual_address(first_section.get_virtual_address());

			//Get reference to last existing PE file
			//section
			const pe_base::section& last_section = image.get_image_sections().back();
			//Calculate total virtual data size
			DWORD total_virtual_size = 
				//Last section virtual address
				last_section.get_virtual_address()
				//Aligned last section virtual size
				+ pe_base::align_up(last_section.get_virtual_size(), image.get_section_alignment())
				//Subtract first section size
				- first_section.get_virtual_address();
			
			//Delete all PE file sections
			image.get_image_sections().clear();

			//Change file alignment if it was more than
			//0x200 - this is minimal allowed value
			//for aligned PE files
			image.realign_file(0x200);

			//Add our section and get reference to
			//already added section with recalculated addresses and sizes
			pe_base::section& added_section = image.add_section(new_section);
			//Set required virtual size for it
			image.set_section_virtual_size(added_section, total_virtual_size);
			
			
			std::cout << "Creating imports..." << std::endl;

			//Create kernel32.dll library imports
			pe_base::import_library kernel32;
			kernel32.set_name("kernel32.dll"); //Set library name

			//Create function to import
			pe_base::imported_function func;
			func.set_name("LoadLibraryA"); //Its name
			kernel32.add_import(func); //Add it to the library

			//And second function
			func.set_name("GetProcAddress");
			kernel32.add_import(func); //Add it, too

			//Get load_library_a field relative address (RVA)
			//of our packed_file_info structure, which was placed at the beginning of 
			//added section, remember?
			DWORD load_library_address_rva = pe_base::rva_from_section_offset(added_section,
			offsetof(packed_file_info, load_library_a));

			//Set this address as
			//import address table (IAT) address
			kernel32.set_rva_to_iat(load_library_address_rva);

			//Create imported library list
			pe_base::imported_functions_list imports;
			//Add our library to it
			imports.push_back(kernel32);

			//Set up import rebuilder
			pe_base::import_rebuilder_settings settings;
			//We don't need original import address table (explanations will be given further) 
			settings.build_original_iat(false);
			//Rewrite IAT to the address
			//which was specified (load_library_address_rva)
			settings.save_iat_and_original_iat_rvas(true, true);
			//Place imports right after packed data end
			settings.set_offset_from_section_start(added_section.get_raw_data().size());
			//Rebuild imports
			image.rebuild_imports(imports, added_section, settings);
		}

		{
			//New section 
			pe_base::section unpacker_section;
			//name - coderpub
			unpacker_section.set_name("coderpub");
			//Available for reading and execution 
			unpacker_section.readable(true).executable(true);
			
			{
				//Get a reference to unpacker section data
				std::string& unpacker_section_data = unpacker_section.get_raw_data();
				//Write unpacker code there
				//This code is stored in autogenerated file
				//unpacker.h, which we included in main.cpp
				unpacker_section_data = std::string(reinterpret_cast<const char*>(unpacker_data), sizeof(unpacker_data));
				//Write image load address 
				//to required offsets
				*reinterpret_cast<DWORD*>(&unpacker_section_data[original_image_base_offset]) = image.get_image_base_32();
				//and first packed file section virtual address,
				//which stores data to unpack and information about them
				//At the beginning of this section, as you remember, 
				//packed_file_info structure is stored
				*reinterpret_cast<DWORD*>(&unpacker_section_data[rva_of_first_section_offset]) = image.get_image_sections().at(0).get_virtual_address();
			}

			//Add this section too
			const pe_base::section& unpacker_added_section = image.add_section(unpacker_section);
			//Set new entry point - now it points to
			//the beginning of unpacker
			image.set_ep(image.rva_from_section_offset(unpacker_added_section, 0));
		}

		//Delete all often used directories
		//we will return them further
		//and manage correctly, but that way for now
		//We leave imports only (and we will not process them yet)
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_BASERELOC);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_EXPORT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_IAT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_RESOURCE);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_SECURITY);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_TLS);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_DEBUG);

		//Strip directory table, removing all empty directories
		//Not completely, but to minimum of 12 elements, because the
		//original file can have first 12 ones in use
		image.strip_data_directories(16 - 4);
		//Remove stub from header, if there was any
		image.strip_stub_overlay();

		
		//Create new PE file
		//Get the name of original file without directory
		std::string base_file_name(argv[1]);
		std::string dir_name;
		std::string::size_type slash_pos;
		if((slash_pos = base_file_name.find_last_of("/\\")) != std::string::npos)
		{
			dir_name = base_file_name.substr(0, slash_pos + 1); //Source file directory
			base_file_name = base_file_name.substr(slash_pos + 1); //Source file name
		}

		//Give a name to a new file: "packed_" + original_file_name
		//Concatenate it with original directory name to save it to a folder where
		//original file is stored
		base_file_name = dir_name + "packed_" + base_file_name;
		//Create file
		std::ofstream new_pe_file(base_file_name.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
		if(!new_pe_file)
		{
			//If failed to create file - display an error message
			std::cout << "Cannot create " << base_file_name << std::endl;
			return -1;
		}

		//Rebuild PE image
		//Strip DOS header, writing NT headers over it
		//(second parameter (true) is responsible for this)
		//Do not recalculate SizeOfHeaders - third parameter is responsible for this
		image.rebuild_pe(new_pe_file, true, false);

		//Message user that file is successfully packed
		std::cout << "Packed image was saved to " << base_file_name << std::endl;
	}
	catch(const pe_exception& e)
	{
		//If the file was not opened by any reason
		//Display error message and exit
		std::cout << e.what() << std::endl;
		return -1;
	}

	return 0;
}
